"""
HandlerSocket backend for sessions in django.
Requires pyhs handlersocket bindings.
"""
import datetime
from django.conf import settings
from django.contrib.sessions.models import Session
from django.contrib.sessions.backends.base import SessionBase, CreateError
from django.core.exceptions import SuspiciousOperation
from django.db import DEFAULT_DB_ALIAS
from django.utils.encoding import force_unicode

from pyhs.manager import Manager
from django.db import models

class SessionStore(SessionBase):
    """
    Implements HandlerSocket session store. It's only valid storage if you are
    using mysql as database engine.
    """
    def __init__(self, session_key=None):
        """
        Initialize pyhs manager, check settings for sanity and pre-cache some
        of them.
        """
        dbs = settings.DATABASES[DEFAULT_DB_ALIAS]
        self._fields_for_index = ['session_key', 'session_data', 'expire_date']
        
        if not 'mysql' in dbs['ENGINE']:
            raise Exception()

        self.db_name = dbs['NAME']
        self.table_name = Session._meta.db_table
        self._session_cache = {}

        self.manager = Manager( [('inet', dbs['HOST'] or 'localhost', 9998)],
                                [('inet', dbs['HOST'] or 'localhost', 9999)])

        super(SessionStore, self).__init__(session_key)

    def load(self):
        """
        Load session data via handlersocket.
        """
        try:
            raw = self.manager.get(self.db_name, self.table_name, 
                    self._fields_for_index, self.session_key)
            
            if raw:
                response = dict(raw)
                
                # low-level check if session is expired
                if datetime.datetime.now() > \
                    models.DateTimeField().to_python(response['expire_date']):
                    raise Session.DoesNotExist()
                
                return self.decode(force_unicode(response['session_data']))
            else:
                raise Session.DoesNotExist()
                
        except (Session.DoesNotExist, SuspiciousOperation):
            self.create()
            return {}

    def exists(self, session_key):
        """
        Check if session exists.
        """
        return bool(self.manager.get(self.db_name, self.table_name, 
                    self._fields_for_index, session_key))

    def create(self):
        """
        Create session.
        """
        while True:
            self.session_key = self._get_new_session_key()
            try:
                # Save immediately to ensure we have a unique entry in the
                # database.
                self.save(must_create=True)
            except CreateError:
                # Key wasn't unique. Try again.
                continue
            self.modified = True
            self._session_cache = {}
            return

    def save(self, must_create=False):
        """
        Save session via handlersocket.
        """
        if self.exists(self.session_key):
            if must_create:
                raise CreateError
                
            self.manager.update(
                self.db_name,
                self.table_name,
                '=', 
                self._fields_for_index,
                [self.session_key],
                [   self.session_key, 
                    self.encode(self._get_session(no_load=must_create)),
                    str(self.get_expiry_date())]
            )
        else:
            self.manager.insert(
                self.db_name,
                self.table_name,
                [   ('session_key', self.session_key), 
                    ('session_data', self.encode(
                        self._get_session(no_load=must_create))), 
                    ('expire_date', str(self.get_expiry_date()))],
            )

    def delete(self, session_key=None):
        """
        Delete session via handlersocket.
        """
        if session_key is None:
            if self._session_key is None:
                return
            session_key = self._session_key

        self.manager.delete(
            self.db_name,
            self.table_name,
            '=', 
            self._fields_for_index,
            [session_key]
        )



DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.mysql', # Add 'postgresql_psycopg2', 'postgresql', 'mysql', 'sqlite3' or 'oracle'.
        'NAME': '',                      # Or path to database file if using sqlite3.
        'USER': '',                      # Not used with sqlite3.
        'PASSWORD': '',                  # Not used with sqlite3.
        'HOST': '',                      # Set to empty string for localhost. Not used with sqlite3.
        'PORT': '',                      # Set to empty string for default. Not used with sqlite3.
    }
}

# Local time zone for this installation. Choices can be found here:
# http://en.wikipedia.org/wiki/List_of_tz_zones_by_name
# although not all choices may be available on all operating systems.
# On Unix systems, a value of None will cause Django to use the same
# timezone as the operating system.
# If running in a Windows environment this must be set to the same as your
# system time zone.
TIME_ZONE = 'America/Chicago'

# Language code for this installation. All choices can be found here:
# http://www.i18nguy.com/unicode/language-identifiers.html
LANGUAGE_CODE = 'en-us'

SITE_ID = 1

# Make this unique, and don't share it with anybody.
SECRET_KEY = '_wri=r&d4&a4&#f+29+ni0@&lzjz!f7luk%)2m@%u74y$8g@-w'

INSTALLED_APPS = (
    'django.contrib.sessions',
    'sessions_test',
)


"""Docstring tests is borrowed from django tests for sessions backend"""

__test__ = {"doctest": """
>>> from django.conf import settings
>>> from hs_sessions.handlersocket import SessionStore as HandlerSocketSession
>>> from django.contrib.sessions.models import Session

>>> db_session = HandlerSocketSession()
>>> db_session.modified
False
>>> db_session.get('cat')
>>> db_session['cat'] = "dog"
>>> db_session.modified
True
>>> db_session.pop('cat')
'dog'
>>> db_session.pop('some key', 'does not exist')
'does not exist'
>>> db_session.save()
>>> db_session.exists(db_session.session_key)
True
>>> db_session.delete(db_session.session_key)
>>> db_session.exists(db_session.session_key)
False

>>> db_session['foo'] = 'bar'
>>> db_session.save()
>>> db_session.exists(db_session.session_key)
True
>>> prev_key = db_session.session_key
>>> db_session.flush()
>>> db_session.exists(prev_key)
False
>>> db_session.session_key == prev_key
False
>>> db_session.modified, db_session.accessed
(True, True)
>>> db_session['a'], db_session['b'] = 'c', 'd'
>>> db_session.save()
>>> prev_key = db_session.session_key
>>> prev_data = db_session.items()
>>> db_session.cycle_key()
>>> db_session.session_key == prev_key
False
>>> db_session.items() == prev_data
True

# Submitting an invalid session key (either by guessing, or if the db has
# removed the key) results in a new key being generated.
>>> Session.objects.filter(pk=db_session.session_key).delete()
>>> db_session = HandlerSocketSession(db_session.session_key)
>>> db_session.save()
>>> HandlerSocketSession('1').get('cat')
"""}




from setuptools import setup

setup(
    name = 'django-hs-sessions',
    version = __import__('hs_sessions').__version__,
    url = 'http://bitbucket.org/dchaplinsky/django-hs-sessions',
    license = 'MIT',
    author = 'Dmitry Chaplinsky',
    author_email='chaplinsky.dmitry@gmail.com',
    packages = ["hs_sessions"],
    install_requires = ['python-handler-socket',],
    description = 'HandlerSocket session backend for django',
    platforms = 'any',

    classifiers = [
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries',
        'Topic :: Database',
        'Framework :: Django',
    ],
)